import numpy as np
import scipy.optimize as opt

def compute_lmn_angles(theta, phi, psi):
    l1 = np.cos(psi) * np.cos(phi) - np.cos(theta) * np.cos(psi) * np.sin(phi)
    m1 = np.sin(psi) * np.cos(phi) + np.cos(theta) * np.cos(psi) * np.sin(phi)
    n1 = np.sin(theta) * np.sin(phi)
    
    l2 = -np.cos(psi) * np.cos(phi) - np.cos(theta) * np.sin(psi) * np.cos(phi)
    m2 = -np.sin(psi) * np.cos(phi) + np.cos(theta) * np.cos(psi) * np.cos(phi)
    n2 = np.sin(theta) * np.cos(phi)
    
    l3 = np.sin(theta) * np.sin(psi)
    m3 = -np.sin(theta) * np.cos(psi)
    n3 = np.cos(theta)
    
    return l1, m1, n1, l2, m2, n2, l3, m3, n3


def max_variance_old(multiaxial_psd, df, method, K=None):
    '''Determine critical plane
    based on max variance.

    Notes
    -----
    Took this from article (its older tough):
        - Lagoda, Macha and Nieslony; Fatigue Life Calculation
        by means of the cycle counting and spectral methods
        under multiaxial random loading.
    '''
    mu = np.sum(multiaxial_psd, axis=0) * df

    if 'maxnormal' == method:

        def crit(v):
            """Criterion for optimization."""
            v = v / np.sqrt(np.sum(v ** 2))
            l, m, n = v
            a = np.asarray([l**2, m**2, n**2,
                            2*l*m, 2*l*n, 2*m*n])

            return -np.abs(np.einsum('i,ij,j', a, mu, a).real)


        res = opt.minimize(fun=crit, x0=[0.33,0.33,0.33], method='SLSQP', bounds=[(0,1),(0,1),(0,1)], options={'disp':False})

    return res['x'] / np.sqrt(np.sum(res['x']**2))


def max_variance(multiaxial_psd, df, method, K=None, search_method='local'):
    '''Determine critical plane
    based on max variance.

    returns: l1, m1, n1, l2, m2, n2, l3, m3, n3
    
        -Lagoda, Macha and Nieslony; Fatigue Life Calculation
        by means of the cycle counting and spectral methods
        under multiaxial random loading.

        -Bȩdkowski, W., and E. Macha; Fatigue fracture plane
        under multiaxial random loadings–prediction by variance
        of equivalent stress based on the maximum shear and normal stresses.
    '''
    mu = np.sum(multiaxial_psd, axis=0) * df

    
    if 'maxnormal' == method:

        def crit(v):
            """Criterion for max-normal."""
            theta, phi, psi = v
            l1, m1, n1, l2, m2, n2, l3, m3, n3 = compute_lmn_angles(theta, phi, psi)
            a = np.asarray([l1**2, m1**2, n1**2,
                            2*l1*m1, 2*l1*n1, 2*m1*n1])


            return -np.abs(np.einsum('i,ij,j', a, mu, a).real)

        if search_method == 'local':
            res = opt.minimize(crit, [np.pi, np.pi, np.pi], method='L-BFGS-B', bounds=[(0, np.pi), (0, 2*np.pi), (0, 2*np.pi)], options={'disp':False})
        
        elif search_method == 'global':
            res = opt.differential_evolution(crit, bounds=[(0, np.pi), (0, 2*np.pi), (0, 2*np.pi)])

        else:
            raise ValueError('Search method') 


    elif 'maxshear' == method:

        def crit(v):
            """Criterion for max-shear."""
            theta, phi, psi = v
            l1, m1, n1, l2, m2, n2, l3, m3, n3 = compute_lmn_angles(theta, phi, psi)
            a = np.asarray([l1**2 - l3**2, m1**2 - m3**2, n1**2 - n3**2,
                            2 * (l1 * m1 - l3 * m3), 2 * (l1 * n1 - l3 * n3), 2 * (m1 * n1 - m3 * n3)])

            return -np.abs(np.einsum('i,ij,j', a, mu, a).real)

        if search_method == 'local':
            res = opt.minimize(crit, [np.pi, np.pi, np.pi], method='L-BFGS-B', bounds=[(0, np.pi), (0, 2*np.pi), (0, 2*np.pi)], options={'disp': False})
        
        elif search_method == 'global':
            res = opt.differential_evolution(crit, bounds=[(0, np.pi), (0, 2*np.pi), (0, 2*np.pi)])

    elif 'maxnormalshear' == method:
        if K is None:
            raise ValueError

        def crit(v):
            """Criterion for max-normal-and-shear."""
            theta, phi, psi = v
            l1, m1, n1, l2, m2, n2, l3, m3, n3 = compute_lmn_angles(theta, phi, psi)
            a = np.asarray([l1**2 - l3**2 + K * (l1 + l3)**2,
                            m1**2 - m3**2 + K * (m1 + m3)**2,
                            n1**2 - n3**2 + K * (n1 + n3)**2,
                            2 * (l1 * m1 - l3 * m3 + K * (l1 + l3) * (m1 + m3)),
                            2 * (l1 * n1 - l3 * n3 + K * (l1 + l3) * (n1 + n3)),
                            2 * (m1 * n1 - m3 * n3 + K * (m1 + m3) * (n1 + n3))]) / (1 + K)

            return -np.abs(np.einsum('i,ij,j', a, mu, a).real)

        if search_method == 'local':
            res = opt.minimize(crit, [np.pi/3, np.pi/3, np.pi/3], method='L-BFGS-B', bounds=[(0, np.pi), (0, 2*np.pi), (0, 2*np.pi)], options={'disp': False})
        
        elif search_method == 'global':
            res = opt.differential_evolution(crit, bounds=[(0, np.pi), (0, 2*np.pi), (0, 2*np.pi)])

    else:
        raise ValueError('Critical plane method')

    direction_cosines = compute_lmn_angles(*res['x'])
    
    return direction_cosines


# Carpinteri-Spagnoli

def C1rot(cpsi, spsi):
    '''Rotation matrix for angle psi.'''
    return np.array([[1, 0, 0, 0, 0, 0],
                     [0, cpsi**2, spsi**2, 2*cpsi*spsi, 0, 0],
                     [0, spsi**2, cpsi**2, -2*cpsi*spsi, 0, 0],
                     [0, -2*cpsi*spsi, 2*cpsi*spsi, cpsi**2-spsi**2, 0, 0],
                     [0, 0, 0, 0, cpsi, -spsi],
                     [0, 0, 0, 0, spsi, -cpsi]])

def C2rot(cth, sth):
    '''Rotation matrix for angle theta.'''
    return np.array([[cth**2, 0, sth**2, 0, 2*cth*sth, 0],
                     [0, 1, 0, 0, 0, 0],
                     [sth**2, 0, cth**2, 0, -2*cth*sth, 0],
                     [0, 0, 0, cth, 0, -sth],
                     [-cth*sth, 0, sth*sth, 0, cth**2-sth**2, 0],
                     [0, 0, 0, sth, 0, cth]])

def C3rot(cphi, sphi):
    '''Rotation matrix for angle phi.'''
    return np.array([[1, 0, 0, 0, 0, 0],
                     [0, cphi**2, sphi**2, 2*cphi*sphi, 0, 0],
                     [0, sphi**2, cphi**2, -2*cphi*sphi, 0, 0],
                     [0, -2*cphi*sphi, 2*cphi*sphi, cphi**2-sphi**2, 0, 0],
                     [0, 0, 0, 0, cphi, -sphi],
                     [0, 0, 0, 0, sphi, cphi]])

def C4rot(cgam, sgam):
    '''Rotation matrix for angle gamma.'''
    return np.array([[cgam**2, sgam**2, 0, 0, 0, 2*cgam*sgam],
                     [sgam**2, cgam**2, 0, 0, 0, -2*cgam*sgam],
                     [0, 0, 1, 0, 0, 0],
                     [0, 0, 0, cgam, sgam, 0],
                     [0, 0, 0, -sgam, cgam, 0],
                     [-cgam*sgam, cgam*sgam, 0, 0, 0, cgam**2-sgam**2]])

def C5rot(cdel, sdel):
    '''Rotation matrix for angle delta.'''
    return np.array([[1, 0, 0, 0, 0, 0],
                     [0, cdel**2, sdel**2, -2*cdel*sdel, 0, 0],
                     [0, sdel**2, cdel**2, 2*cdel*sdel, 0, 0],
                     [0, cdel*sdel, -cdel*sdel, 0, 0, 0],
                     [0, 0, 0, 0, cdel, sdel],
                     [0, 0, 0, 0, -sdel, cdel]])


def spectral_moment(f, multiaxial_psd, n):
    """Calculate n-th spectral moment of a PSD.

    f: frequency array
    s: PSD array
    n: order of spectral moment
    
    Returns
    -------
    Scalar.
    """
    df = f[1] - f[0]
    
    #return np.tensordot(f**n, s, ([0], [0])) * df
    return np.einsum('i,i...->...', f**n, multiaxial_psd) * df


def csrandom(multiaxial_psd, df, s_af, tau_af):
    '''Determine critical plane with
    the C-S random criterion.
    
    See also
    --------
    Carpinteri A, Spagnoli A and Vantadori S, Reformulation in the frequency domain of a critical
    plane-based multiaxial fatigue criterion, Int J Fat, 2014
    '''

    f = np.arange(0, multiaxial_psd.shape[0]*df, df)
    l0 = spectral_moment(f, multiaxial_psd, 0)
    N1 = np.sqrt(spectral_moment(f, multiaxial_psd, 2) / l0) / (2*np.pi)
    
    # -- S33 first; extreme according to Davenport.
    # .. Varying angles 0 < phi < 2pi and 0 < theta < pi
    def s33crit(inp):
        """Optimisation criterion (searching
        for maximum)."""
        phi, th = inp
        C = np.dot(C2rot(np.cos(th), np.sin(th)), C3rot(np.cos(phi), np.sin(phi)))
        l0_ = np.dot(np.dot(C, l0), C.transpose())
        N1_ = np.dot(np.dot(C, N1), C.transpose())
        
        return -(np.sqrt(l0_[2,2]) * np.sqrt(2*np.log(N1_[2,2]))
                 + 0.5772/np.sqrt(2*np.log(N1_[2,2])))
        
    res = opt.minimize(s33crit, [0.33,0.33], method='SLSQP', bounds=[(0,2*np.pi),(0,np.pi)], options={'disp':False})

    #res = opt.differential_evolution(s33crit, bounds=[(0,2*np.pi),(0,np.pi)])

    


    phi, th = res['x']
    
    # -- Tau_yz second; maximum variance.
    def taucrit(psi):
        '''Optimisation criterion (searching for
        maximum variance (mu).'''
        C = np.dot(
                   np.dot(C1rot(np.cos(psi), np.sin(psi)), C2rot(np.cos(th), np.sin(th))),
                   C3rot(np.cos(phi), np.sin(phi)))
        
        return -np.dot(np.dot(C, l0), C.transpose())[-1, -1]
        
    
    res = opt.minimize_scalar(taucrit, 0.33, method='bounded', bounds=(0,2*np.pi), options={'disp':False})
    #res = opt.differential_evolution(taucrit, bounds=(0,2*np.pi))
    psi = res['x'].real
    

    delta = 3*np.pi*(1-(tau_af)**2/(s_af)**2)/8

    # -- Maximize shear variance
    C = np.dot(
               np.dot(C1rot(np.cos(psi), np.sin(psi)), C2rot(np.cos(th), np.sin(th))),
               C3rot(np.cos(phi), np.sin(phi)))
    def tau2crit(gam):
        '''Optimisation criterion (searching for
        maximum variance).'''
        CC = np.dot(
                    np.dot(C4rot(np.cos(gam), np.sin(gam)), C5rot(np.cos(delta), np.sin(delta))), C)
        
        return -np.dot(np.dot(CC, l0), C.transpose())[-1, -1]
    
    res = opt.minimize_scalar(tau2crit, 0.33, method='bounded', bounds=(0, 2*np.pi), options={'disp':False})
    #res = opt.differential_evolution(tau2crit, bounds=(0,2*np.pi))
    gam = res['x'].real
    
    CC = np.dot(
            np.dot(C4rot(np.cos(gam), np.sin(gam)), C5rot(np.cos(delta), np.sin(delta))), C)
    
    return CC






import numpy as np
from. narrowband import Narrowband

class HuangMoan(Narrowband):
    """Class for fatigue life estimation using frequency domain 
    method by Huang and Moan[1].
      
    References
    ----------
    [1] Wenbo Huang and Torgeir Moan. Fatigue Under Combined High and Low Frequency Loads.
        25th International Conference on Offshore Mechanics and Arctic Engineering,
        Hamburg, Germany, 2006. ASME, Paper No. OMAE2006–92247.
    [2] Aleš Zorman and Janko Slavič and Miha Boltežar. 
        Vibration fatigue by spectral methods—A review with open-source support, 
        Mechanical Systems and Signal Processing, 2023, 
        https://doi.org/10.1016/j.ymssp.2023.110149
    
    Example
    -------
    Import modules, define time- and frequency-domain data

    >>> import FLife
    >>> import pyExSi as es
    >>> import numpy as np
    >>> from matplotlib import pyplot as plt
    >>> # time-domain data
    >>> N = 2 ** 16  # number of data points of time signal
    >>> fs = 2048  # sampling frequency [Hz]
    >>> t = np.arange(0, N) / fs  # time vector
    >>> # frequency-domain data
    >>> M = N // 2 + 1  # number of data points of frequency vector
    >>> freq = np.arange(0, M, 1) * fs / N  # frequency vector
    >>> PSD_lower = es.get_psd(freq, 20, 60, variance = 5)  # lower mode of random process
    >>> PSD_higher = es.get_psd(freq, 100, 120, variance = 2)  # higher mode of random process
    >>> PSD = PSD_lower + PSD_higher # bimodal one-sided flat-shaped PSD

    Get Gaussian stationary signal, instantiate SpectralData object and plot PSD

    >>> rg = np.random.default_rng(123) # random generator seed
    >>> x = es.random_gaussian(N, PSD, fs, rg) # Gaussian stationary signal
    >>> sd = FLife.SpectralData(input=x, dt=1/fs) # SpectralData instance
    >>> plt.plot(sd.psd[:,0], sd.psd[:,1]) 
    >>> plt.xlabel('Frequency [Hz]')
    >>> plt.ylabel('PSD')

    Define S-N curve parameters and get fatigue-life estimatate

    >>> C = 1.8e+22  # S-N curve intercept [MPa**k]
    >>> k = 7.3 # S-N curve inverse slope [/]
    >>> hm = FLife.HuangMoan(sd, PSD_splitting=('userDefinedBands', [80,150]))
    >>> print(f'Fatigue life: {hm.get_life(C,k):.3e} s.')

    Plot segmentated PSD, used in HuangMoan method

    >>> lower_band_index, upper_band_index= hm.band_stop_indexes
    >>> plt.plot(sd.psd[:,0], sd.psd[:,1])
    >>> plt.vlines(sd.psd[:,0][lower_band_index], 0, np.max(sd.psd[:,1]), 'k', linestyles='dashed', alpha=.5)
    >>> plt.fill_between(sd.psd[:lower_band_index,0], sd.psd[:lower_band_index,1], 'o', label='lower band', alpha=.2, color='blue')
    >>> plt.vlines(sd.psd[:,0][upper_band_index], 0, np.max(sd.psd[:,1]), 'k', linestyles='dashed', alpha=.5)
    >>> plt.fill_between(sd.psd[lower_band_index:upper_band_index,0], sd.psd[lower_band_index:upper_band_index,1], 'o', label='upper band', alpha=.5, color ='orange')
    >>> plt.xlabel('Frequency [Hz]')
    >>> plt.ylabel('PSD')
    >>> plt.xlim(0,300)
    >>> plt.legend()
    """
    def __init__(self, spectral_data, PSD_splitting = ('equalAreaBands', 2)):
        """Get needed values from reference object.

        :param spectral_data: Instance of class SpectralData
        :param PSD_splitting: tuple
                PSD_splitting[0] is PSD spliting method, PSD_splitting[1] is method argument. 
                Splitting methods:

                - 'userDefinedBands', PSD_splitting[1] must be of type list or tupple, with N 
                  elements specifying upper band frequencies of N random processes.
                - 'equalAreaBands', PSD_splitting[1] must be of type int, specifying N random processes.

                Defaults to ('equalAreaBands', 2).
        """ 
        Narrowband.__init__(self, spectral_data)
        self.PSD_splitting = PSD_splitting
        self.band_stop_indexes = self.spectral_data._get_band_stop_frequency(self.PSD_splitting)

    def get_life(self, C, k):
        """Calculate fatigue life with parameters C, k, as defined in [2].
        
        :param C: [int,float]
            S-N curve intercept [MPa**k].
        :param k: [int,float]
            S-N curve inverse slope [/].
        :return:
            Estimated fatigue life in seconds.
        :rtype: float
        """
        moments = self.spectral_data.get_spectral_moments(self.PSD_splitting, moments=[0])
        m0L, = moments[0] #spectral moments for lower band
        m0H, = moments[1] #spectral moments for upper band

        # -- positive slope zero crossing frequency
        v0L, v0H = self.spectral_data.get_nup(self.PSD_splitting)

        # -- damage intensity
        dNB_H = self.damage_intesity_NB(m0=m0H, nu=v0H, C=C, k=k)
        dNB_L = self.damage_intesity_NB(m0=m0L, nu=v0L, C=C, k=k) 

        d =  ((dNB_L/v0L)**(2/k) + (dNB_H/v0H)**(2/k))**((k-2)/2) \
            * (v0L**2 * (dNB_L/v0L)**(2/k) + v0H**2 * (dNB_H/v0H)**(2/k) )**(3/2) \
                        / ( v0L**4 * (dNB_L/v0L)**(2/k) +  v0H**4 * (dNB_H/v0H)**(2/k) )**(1/2)
        T= 1/d
        
        return T

    def get_PDF(self, s):
        raise Exception(f'Function <get_PDF> is not available for class {self.__class__.__name__:s}.')


import numpy as np
from . import cplane
from . import criteria
from ..spectralData import SpectralData
from scipy.optimize import minimize

class EquivalentStress(SpectralData): #equivalentstress
    """EquivalentStress object contains multiaxial data, based on input power spectral density (PSD).
    It is used to convert multiaxial stress states into equivalent uniaxiail stress states, that can be used in fatigue life estimation.

    EquivalentStress inherits from SpectralData class, so all methods from SpectralData are available for fatigue analysis.
    If instance of SpectralData is passed as input, the multiaxial PSD is inherited, otherwise it is created from input PSD.
    
    -----
    
    The following multiaxial criteria are available for equivalent stress calculation:
    - max_normal: Maximum normal stress on critical plane
    - max_shear: Maximum shear stress on critical plane
    - max_normal_and_shear: Maximum normal and shear stress on critical plane
    - EVMS: Equivalent von Misses stress
    - cs: Carpinteri-Spagnoli criterion
    - multiaxial_rainflow: Frequency-based multiaxial rainflow criterion
    - thermoelastic: Thermoelasticity-based criterion
    - EVMS_out_of_phase: EVMS adaptation for out-of-phase components
    - Nieslony: Nieslony criterion, combining von Mises and hydrostatic stresses
    - Lemaitre: Equivalent Lemaitre stress
    - liwi: LiWI approach
    - coin_liwi: COIN-LiWI method

    """

    def __init__(self, input=None, window='hann', nperseg=1280,
                 noverlap=None, psd_trim_length=None, 
                 T=None, fs=None, rg=None,
                 **kwargs):
        """
        Class constructor for EquivalentStress object.

        :param input: dictionary or instance of SpectralData class.
                Input data for EquivalentStress object. If dictionary, it should contain keys: `PSD` and `f`
                or keys 'amplitude_spectrum' and 'f'.
                PSD should be an array with shape (f,6,6) for 3D stress state or (f,3,3) for 2D stress state.
                If multi-point PSD is provided, the shape should be (N,f,6,6) or (N,f,3,3) respectively.
                Amplitude spectrum should be an array with shape (f,3) for 2D stress state or (f,6) for 3D stress state.
                If multi-point amplitude spectrum is provided, the shape should be (N,f,3) or (N,f,6) respectively.
                If instance of SpectralData class is passed, the multiaxial PSD is inherited.
        :param window: str or tuple or array_like, optional
                Desired window to use. Defaults to ‘hann’.
        :param nperseg:  int, optional
                Length of each segment. Defaults to 1280.
        :param noverlap: int, optional
                Number of points to overlap between segments. If None, 
                noverlap = nperseg / 2. Defaults to None.
        :param psd_trim_length: int, optional
                Number of frequency points to be used for PSD.
                Defaults to None.
        :param T: int, float, optional
                Length of time history when random process is defined by `input` parameter 'GUI'
                or by PSD and frequency vector. If T and fs are provided, time histoy is generated.
                Defaults to None.
        :param fs: int, float, optional
                Sampling frequency of time history when random process is defined by `input` parameter 
                'GUI' or by PSD and frequency vector. If T and fs are provided, time histoy is generated.
                Defaults to None.
        :param rg: numpy.random._generator.Generator, optional
                Random generator controls phase of generated time history, when `input` is 'GUI' or 
                (PSD, frequency vector).
                Defaults to None.

        """
        # Class in instantiated with EquivalentStress(input), input is dictionary or tuple (PSD,freq)
        if isinstance(input, dict) or isinstance(input, tuple):
            SpectralData.__init__(self,input=input)
            if T is not None and fs is not None:
                self.t = T
                self.fs = fs
        # Class instance is instantiated with SpectralData(input) and inherited with EquivalentStress(spectral_data). Input is an instance of SpectralData class
        elif isinstance(input, SpectralData):
            self.spectral_data = input
            if hasattr(input,'multiaxial_psd'):
                self.multiaxial_psd = self.spectral_data.multiaxial_psd
            elif hasattr(input,'multiaxial_amplitude_spectrum'):
                self.multiaxial_amplitude_spectrum = self.spectral_data.multiaxial_amplitude_spectrum
            if hasattr(input,'t') and hasattr(input,'fs'):
                self.t = input.t
                self.fs = input.fs
            self.multipoint = self.spectral_data.multipoint
        else:
            raise Exception('Unrecognized Input Error. `input` should be a dictionary with keys: `PSD` and `f`).')


    # def is_multiaxial_psd_4d(self):
    #     """Check if the multiaxial PSD is 4D (i.e., multiple points on the model)"""
    #     return len(self.multiaxial_psd[0].shape) == 4


    def loop_over_points(self, criterion, *args, **kwargs):
        """Loop the selected criterion over multiple points on the model"""    
        if hasattr(self,'multiaxial_psd'):
            s_eq_multi_point = np.empty((self.multiaxial_psd[0].shape[:2]))
        elif hasattr(self,'multiaxial_amplitude_spectrum'):
            s_eq_multi_point = np.empty((self.multiaxial_amplitude_spectrum[0].shape[:2]),dtype=complex)
        
        for i in range(len(s_eq_multi_point)):
            if hasattr(self,'multiaxial_psd'):
                s_eq = criterion(self,s=self.multiaxial_psd[0][i],*args, **kwargs)
            elif hasattr(self,'multiaxial_amplitude_spectrum'):
                s_eq = criterion(self,s=self.multiaxial_amplitude_spectrum[0][i],*args, **kwargs)
            s_eq_multi_point[i] = s_eq
        if hasattr(self,'multiaxial_psd'):
            self.eq_psd_multipoint = (s_eq_multi_point, self.multiaxial_psd[1]) #(multipoint psd, freq)
        elif hasattr(self,'multiaxial_amplitude_spectrum'):
            #calculate psd from amplitude spectrum at each point at each frequency
            self.eq_psd_multipoint = (abs(s_eq_multi_point**2), self.multiaxial_amplitude_spectrum[1])

    def set_eq_stress(self,eq_psd,f):
        """Set equivalent stress to the object. Also generate im ehistory if T and fs are provided"""
        self.psd = np.column_stack((f, eq_psd))
        # needed parameters for time-history generation
        if hasattr(self,'t') and hasattr(self, 'fs'):
            self._set_time_history(f=f, psd=eq_psd, T=self.t, fs=self.fs)
        if hasattr(self,'psd'):
            self.PSD_splitting = ('equalAreaBands', 1) 
            self._calculate_coefficients()

    def select_critical_point(self, point_index):
        """Manually select a critical point from multipoint equivalent stress
        
        :param point_index: int
                Index of the critical point

        """
        self.set_eq_stress(eq_psd=self.eq_psd_multipoint[0][point_index], f=self.eq_psd_multipoint[1])

    def max_normal(self,search_method='local'): #max_normal
        """
        Converts the stress tensor at one node or multiple nodes to equivalent,
        scalar psd stress, using method of maximum normal stress.
        
        :param search_method: str, optional, default 'local'
                Search method for optimization. Options are 'local' or 'global'. Local is prefered, unless the optimization fails.
        
        --------

        - Nieslony, Adam and Macha, Ewald (2007);
            Spectral method in multiaxial random fatigue

        """
        # psd is multiple-point
        if self.multipoint: 
            self.loop_over_points(criterion=criteria._max_normal,search_method = 'local')
        
        # psd in single-point
        else: 
            s_eq = criteria._max_normal(self,s=self.multiaxial_psd[0],search_method=search_method)
            self.set_eq_stress(eq_psd=s_eq, f=self.multiaxial_psd[1])


    def max_shear(self,search_method='local'):
        """
        Converts the stress tensor at one node or multiple nodes to equivalent,
        scalar psd stress, using method of maximum shear stress.
        
        :param search_method: str, optional, default 'local'
                Search method for optimization. Options are 'local' or 'global'. Local is prefered, unless the optimization fails.
        
        --------
            
        - Nieslony, Adam and Macha, Ewald (2007);
            Spectral method in multiaxial random fatigue

        """
        # psd is multiple-point
        if self.multipoint:
            self.loop_over_points(criterion=criteria._max_shear,search_method = 'local')
        
        # psd is single-point
        else:
            s_eq = criteria._max_shear(self,s=self.multiaxial_psd[0],search_method=search_method)
            self.set_eq_stress(eq_psd=s_eq, f=self.multiaxial_psd[1])

    def max_normal_and_shear(self, s_af, tau_af, search_method='local'):
        """
        Converts the stress tensor at one node or multiple nodes to equivalent,
        scalar psd stress, using method of maximum normal and shear stress.
        
        Critical plane is based on max variance of shear stress
        
        :param s_af: float
            Fully reversed torsion-fatigue limit. Used for calculating material fatigue coefficient K [1].
        :param tau_af: float
            Fully reversed torsion-fatigue limit. Used Used for calculating material fatigue coefficient K [1].
        
        --------
            
        - Nieslony, Adam and Macha, Ewald (2007);
            Spectral method in multiaxial random fatigue

        - Matjaz Mrsnik, Janko Slavic and Miha Boltezar
            Multiaxial Vibration Fatigue A Theoretical and Experimental Comparison.
            Mechanical Systems and Signal Processing, 2016

        """
        # psd is multiple-point
        if self.multipoint:
            self.loop_over_points(criterion=criteria._max_normal_and_shear, search_method='local', s_af=s_af, tau_af=tau_af)
        
        # psd is single-point
        else:
            s_eq = criteria._max_normal_and_shear(self,s=self.multiaxial_psd[0], s_af=s_af, tau_af=tau_af, search_method=search_method)
            self.set_eq_stress(eq_psd=s_eq, f=self.multiaxial_psd[1])
    
    def EVMS(self):
        """
        Converts the stress tensor at one node or multiple nodes to equivalent,
        scalar psd stress, using ther equivalent von Misses stress in frequency domain (EVMS)
    
        --------

        - Preumont, A., & Piéfort, V. (1994);
            Predicting Random High-Cycle Fatigue Life With Finite Elements.

        """
        # psd is multiple-point
        if self.multipoint:
            self.loop_over_points(criterion=criteria._EVMS)
        
        # psd is single-point
        else:
            s_eq = criteria._EVMS(self,s=self.multiaxial_psd[0])
            self.set_eq_stress(eq_psd=s_eq, f=self.multiaxial_psd[1])

    def cs(self, s_af, tau_af):
        """Converts the stress tensor at one node to equivalent,
        scalar psd stress, using the C-S criterion.
        
        :param s_af: float
            Fully reversed torsion-fatigue limit.
        :param tau_af: float
            Fully reversed torsion-fatigue limit.
        
        --------

        - Carpinteri A, Spagnoli A and Vantadori S; 
            Reformulation in the frequency domain of a critical plane-based multiaxial fatigue criterion,
            Int J Fat, 2014

        """
        # psd is multiple-point
        if self.multipoint:
            self.loop_over_points(criterion=criteria._cs, s_af=s_af, tau_af=tau_af)
        
        # psd is single-point
        else:
            s_eq = criteria._cs(self,s=self.multiaxial_psd[0], s_af=s_af, tau_af=tau_af)
            self.set_eq_stress(eq_psd=s_eq, f=self.multiaxial_psd[1])

    def multiaxial_rainflow(self):
        """Converts the stress tensor at one node to equivalent,
        scalar psd stress, for use in frequency domain multiaxial rainflow.

        ONLY WORKS FOR BIAXIAL STRESSES: PSD MATRIX (f,3,3) - single point or (N,f,3,3) - multiple points
    
        --------

        - Pitoiset, Xavier, and André Preumont; 
            Spectral methods for multiaxial random fatigue analysis of metallic structures, 
            International journal of fatigue, 2000

        """
        # psd is multiple-point
        if self.multipoint:
            self.loop_over_points(criterion=criteria._multiaxial_rainflow)
        
        # psd is single-point
        else:
            s_eq = criteria._multiaxial_rainflow(self,s=self.multiaxial_psd[0])
            self.set_eq_stress(eq_psd=s_eq, f=self.multiaxial_psd[1])

    def thermoelastic(self):
        """Converts the stress tensor at one node to equivalent,
        scalar psd stress, using the thermoelasticity based criterion, 
        simulating the equivalent stress, detected with the thermal camera

        --------

        - Šonc J, Zaletelj K and Slavič J;
            Application of thermoelasticity in the frequency-domain multiaxial vibration-fatigue criterion, 
            Mechanical Systems and Signal Processsing, 2025

        """
        # psd is multiple-point
        if self.multipoint:
            self.loop_over_points(criterion=criteria._thermoelastic)
        
        # psd is single-point
        else:
            s_eq = criteria._thermoelastic(self,s=self.multiaxial_psd[0])
            self.set_eq_stress(eq_psd=s_eq, f=self.multiaxial_psd[1])

    def liwi(self):
        """Converts the stress tensor at one node to equivalent
        stress, using the LiWI approach.

        ONLY WORKS WITH BIAXIAL AMPLITUDE SPECTRUM (f,3) - single point or (N,f,3) - multiple points

        --------

        - Alexander T. Schmidt, Nimish Pandiya,
            Extension of the static equivalent stress hypotheses to linearly vibrating systems using wave interference – The LiWI approach,
            International Journal of Fatigue, 2021,

        """
        if not hasattr(self,'multiaxial_amplitude_spectrum'):
            raise Exception('Input Error. LiWI approach only works with multiaxial amplitude spectrum.')
        
        # psd is multiple-point
        if self.multipoint:
            self.loop_over_points(criterion=criteria._liwi)
        
        # psd is single-point
        else:
            s_eq_amplitude_spectrum = criteria._liwi(self,s=self.multiaxial_amplitude_spectrum[0])

            for i in range(len(s_eq_amplitude_spectrum)):
                s_eq_psd = np.dot(s_eq_amplitude_spectrum[i], np.conj(s_eq_amplitude_spectrum[i].T))

            self.set_eq_stress(eq_psd=s_eq_psd, f=self.multiaxial_psd[1])


    def coin_liwi(self, k_a, k_phi):
        """Converts the stress tensor at one node to equivalent
        stress, using the COIN-LiWI method.

        ONLY WORKS WITH 3D AMPLITUDE SPECTRUM (f,6) - single point or (N,f,6) - multiple points
        
        :param k_a: float
            Tension shear strength ratio. (from article: 1.70 for aluminum alloy, 1.64 for structural steel, 1.43 for cast iron)
        :param k_phi: float
            Phase influence factor (from article: 0.90 for aluminum alloy, 0.85 for structural steel, 1.10 for cast iron)
        
        --------

        - Alexander T. Schmidt, Jan Kraft,
            A new equivalent stress approach based on complex invariants: The COIN LiWI method,
            International Journal of Fatigue, 2023
        
        """
        if not hasattr(self,'multiaxial_amplitude_spectrum'):
            raise Exception('Input Error. COIN LiWI method only works with multiaxial amplitude spectrum.')
        
        # psd is multiple-point
        if self.multipoint:
            self.loop_over_points(criterion=criteria._coin_liwi, k_a=k_a, k_phi=k_phi)
        
        # psd is single-point
        else:
            s_eq_amplitude_spectrum = criteria._coin_liwi(self,s=self.multiaxial_amplitude_spectrum[0], k_a=k_a, k_phi=k_phi)

            for i in range(len(s_eq_amplitude_spectrum)):
                s_eq_psd = np.dot(s_eq_amplitude_spectrum[i], np.conj(s_eq_amplitude_spectrum[i].T))

            self.set_eq_stress(eq_psd=s_eq_psd, f=self.multiaxial_psd[1])


    def EVMS_out_of_phase(self):
        """Converts the stress tensor at one node to equivalent
        stress, using the phase-sensitive EVMS criterion, adjusted for out-of-phase components.

        ONLY WORKS WITH BIAXIAL PSD (f,3,3) - single point or (N,f,3,3) - multiple points
               
        --------

        - Bonte, Martijn HA and de Boer, Andries and Liebregts R
            Determining the von Mises stress power spectral density for frequency domain fatigue analysis including out-of-phase stress components,
            Journal of Sound and Vibration, 2007
        
        """
        
        # psd is multiple-point
        if self.multipoint:
            self.loop_over_points(criterion=criteria._EVMS_out_of_phase)
        
        # psd is single-point
        else:
            s_eq_psd = criteria._EVMS_out_of_phase(self,s=self.multiaxial_psd[0])

            self.set_eq_stress(eq_psd=s_eq_psd, f=self.multiaxial_psd[1])


    def Nieslony(self, s_af, tau_af, coefficient_load_type='tension'):
        """Converts the stress tensor at one node to equivalent,
        scalar psd stress, using the Nieaslony criterion, combining von Mises and hydrostatic stresses.

        :param s_af: float
            Fully reversed torsion-fatigue limit.
        :param tau_af: float
            Fully reversed torsion-fatigue limit.
        :param coefficient_load_type: str
            Type of load with which the material fatigue properties were determined. Options are 'tension' for tension-compression load or 'torsion'.
        
        --------

        - Adam Niesłony, Michał Böhm, Robert Owsiński, Artur Dziura, Karol Czekaj,
            Integrating von Mises and hydrostatic stresses in frequency domain multiaxial fatigue criteria for vibration fatigue analysis,
            Mechanical Systems and Signal Processing, 2025

        """
        # psd is multiple-point
        if self.multipoint:
            self.loop_over_points(criterion=criteria._Nieslony, s_af=s_af, tau_af=tau_af, coefficient_load_type=coefficient_load_type)
        
        # psd is single-point
        else:
            s_eq = criteria._Nieslony(self,s=self.multiaxial_psd[0], s_af=s_af, tau_af=tau_af, coefficient_load_type=coefficient_load_type)
            self.set_eq_stress(eq_psd=s_eq, f=self.multiaxial_psd[1])

    def Lemaitre(self, poisson_ratio):
        """
        Converts the stress tensor at one node or multiple nodes to equivalent,
        scalar psd stress, using ther equivalent Lemaitre stress in frequency domain
    
        --------

        - Jingran Ge, Yi Sun, Song Zhou;
            Fatigue life estimation under multiaxial random loading by means of the equivalent Lemaitre stress and multiaxial S–N curve methods.
            International Journal of Fatigue, 2015
            
        """
        # psd is multiple-point
        if self.multipoint:
            self.loop_over_points(criterion=criteria._Lemaitre, poisson_ratio=poisson_ratio)
        
        # psd is single-point
        else:
            s_eq = criteria._Lemaitre(self,s=self.multiaxial_psd[0], poisson_ratio=poisson_ratio)
            self.set_eq_stress(eq_psd=s_eq, f=self.multiaxial_psd[1])


import numpy as np
from scipy import stats
from scipy import integrate
from .jiao_moan import JiaoMoan
from ..tools import pdf_rayleigh_sum

class GaoMoan(JiaoMoan):
    """Class for fatigue life estimation using frequency domain 
    method by Gao and Moan [1].
    
    References
    ----------
    [1] Zhen Gao and Torgeir Moan. Frequency-domain fatigue analysis of
        wide-band stationary Gaussian processes using a trimodal spectral formulation.
        International Journal of Fatigue, 30(10-11): 1944-1955, 2008
    [2] Aleš Zorman and Janko Slavič and Miha Boltežar. 
        Vibration fatigue by spectral methods—A review with open-source support, 
        Mechanical Systems and Signal Processing, 2023, 
        https://doi.org/10.1016/j.ymssp.2023.110149
    
    Example
    -------
    Import modules, define time- and frequency-domain data

    >>> import FLife
    >>> import pyExSi as es
    >>> import numpy as np
    >>> from matplotlib import pyplot as plt
    >>> # time-domain data
    >>> N = 2 ** 16  # number of data points of time signal
    >>> fs = 2048  # sampling frequency [Hz]
    >>> t = np.arange(0, N) / fs  # time vector
    >>> # frequency-domain data
    >>> M = N // 2 + 1  # number of data points of frequency vector
    >>> freq = np.arange(0, M, 1) * fs / N  # frequency vector
    >>> PSD_lower = es.get_psd(freq, 20, 60, variance = 5)  # lower mode of random process
    >>> PSD_middle = es.get_psd(freq, 100, 120, variance = 1)  # middle mode of random process
    >>> PSD_higher = es.get_psd(freq, 300, 350, variance = 2)  # higher mode of random process
    >>> PSD = PSD_lower + PSD_middle + PSDb_higher # trimodal one-sided flat-shaped PSD

    Get Gaussian stationary signal, instantiate SpectralData object and plot PSD

    >>> rg = np.random.default_rng(123) # random generator seed
    >>> x = es.random_gaussian(N, PSD, fs, rg) # Gaussian stationary signal
    >>> sd = FLife.SpectralData(input=x, dt=1/fs) # SpectralData instance
    >>> plt.plot(sd.psd[:,0], sd.psd[:,1]) 
    >>> plt.xlabel('Frequency [Hz]')
    >>> plt.ylabel('PSD')

    Define S-N curve parameters and get fatigue-life estimatate

    >>> C = 1.8e+22  # S-N curve intercept [MPa**k]
    >>> k = 7.3 # S-N curve inverse slope [/]
    >>> gm = FLife.GaoMoan(sd, PSD_splitting=('userDefinedBands', [80,150,400]))  # fatigue-life estimator
    >>> print(f'Fatigue life: {gm.get_life(C,k):.3e} s.')   

    Plot segmentated PSD, used in Gao-Moan method

    >>> lower_band_index, middle_band_index, upper_band_index= gm.band_stop_indexes
    >>> plt.plot(sd.psd[:,0], sd.psd[:,1])
    >>> plt.vlines(sd.psd[:,0][lower_band_index], 0, np.max(sd.psd[:,1]), 'k', linestyles='dashed', alpha=.5)
    >>> # lower band
    >>> plt.fill_between(sd.psd[:lower_band_index,0], sd.psd[:lower_band_index,1], 'o', label='lower band', alpha=.2, color='blue')
    >>> plt.vlines(sd.psd[:,0][middle_band_index], 0, np.max(sd.psd[:,1]), 'k', linestyles='dashed', alpha=.5)
    >>> # middle band
    >>> plt.fill_between(sd.psd[lower_band_index:middle_band_index,0], sd.psd[lower_band_index:middle_band_index,1], 'o', label='middle band', alpha=.5, color ='orange')
    >>> plt.vlines(sd.psd[:,0][upper_band_index], 0, np.max(sd.psd[:,1]), 'k', linestyles='dashed', alpha=.5)
    >>> # upper band
    >>> plt.fill_between(sd.psd[middle_band_index:upper_band_index,0], sd.psd[middle_band_index:upper_band_index,1], 'o', label='upper band', alpha=.5, color ='green')
    >>> plt.xlabel('Frequency [Hz]')
    >>> plt.ylabel('PSD')
    >>> plt.xlim(0,450)
    >>> plt.legend()
    """
    def __init__(self, spectral_data, PSD_splitting = ('equalAreaBands', 3)):
        """Get needed values from reference object.

        :param spectral_data:  Instance of class SpectralData       
        :param PSD_splitting: tuple
                PSD_splitting[0] is PSD spliting method, PSD_splitting[1] is method argument. 
                Splitting methods:

                - 'userDefinedBands', PSD_splitting[1] must be of type list or tupple, with N 
                  elements specifying upper band frequencies of N random processes.
                - 'equalAreaBands', PSD_splitting[1] must be of type int, specifying N random processes.

                Defaults to ('equalAreaBands', 3).
        """
        JiaoMoan.__init__(self, spectral_data, PSD_splitting)
        
    def get_life(self, C, k):
        """Calculate fatigue life with parameters C, k, as defined in [1, 2].

        :param C: [int,float];
            S-N curve intercept [MPa**k].
        :param k: [int,float];
            S-N curve inverse slope [/].
        :return:
            Estimated fatigue life in seconds.
        :rtype: float
        """
        if len(self.band_stop_indexes) == 1: # narrow-band
            T = self._life_NB(C,k)
        elif len(self.band_stop_indexes) == 2: # bi-modal
            T = self._life_bimodal(C,k)
        elif len(self.band_stop_indexes) == 3: # tri-modal
            T = self._life_trimodal(C,k)
        else:
            raise Exception('Specify up to tri-modal random process.')
        return T

    def _life_bimodal(self, C, k):
        # -- spectral moments for each narrowband
        moments = self.spectral_data.get_spectral_moments(self.PSD_splitting, moments=[0,2])
        m0L, m2L = moments[0] #spectral moments for lower band
        m0H, m2H = moments[1] #spectral moments for upper band

        # -- Vanmarcke bandwidth parameter
        _, epsV_H = self.spectral_data.get_vanmarcke_parameter(self.PSD_splitting)

        # -- positive slope zero crossing frequency
        _, v0H = self.spectral_data.get_nup(self.PSD_splitting)
        v0P = 1/(2*np.pi) * (np.sqrt(m0L) / (m0H + m0L)) * np.sqrt(m2H * epsV_H**2 + m2L) 
        #v0P1 = m0L_norm * v0L* np.sqrt(1 + m0H_norm/m0L_norm * (v0H/v0L*epsV_H)**2) #jiao-moan, izraz je ekvivalenten za bimodalen proces

        # -- damage intensity
        dNB_H = self.damage_intesity_NB(m0=m0H, nu=v0H, C=C, k=k) 
        dNB_P = self._damage_intesity_bimodal_LF(m0L=m0L, m0H=m0H, nuP=v0P, C=C, k=k)  #deduje od jiao-moana

        d = dNB_H + dNB_P
        T = 1 / d
        return T

    def _life_trimodal(self, C, k):
        moments = self.spectral_data.get_spectral_moments(self.PSD_splitting, moments=[0,2])
        m0L, m2L = moments[0]
        m0M, m2M = moments[1]
        m0H, m2H = moments[2]

        # -- Vanmarcke bandwidth parameter
        _, epsV_M, epsV_H = self.spectral_data.get_vanmarcke_parameter(self.PSD_splitting)

        # -- positive slope zero crossing frequency
        _, _, v0H = self.spectral_data.get_nup(self.PSD_splitting)

        # -- positive slope zero crossing frequency
        # -- process HF + MF
        v0P = 1/(2*np.pi) * (np.sqrt(m0M) / (m0H + m0M)) * np.sqrt(m2H * epsV_H**2 + m2M) 

        # -- process HF + MF + LF
        v0Q = 1/(4*np.pi) * np.sqrt(m2H * epsV_H**2 + m2M * epsV_M**2 + m2L) * \
            (2.0 * np.sqrt(m0L * (m0H + m0M + m0L)) - np.pi * np.sqrt(m0H * m0M) \
            + 2.0 * np.sqrt(m0H * m0M) * np.arctan(np.sqrt((m0H * m0M ) /m0L) / np.sqrt(m0H + m0M + m0L))) \
            / (np.sqrt(m0H + m0M + m0L)**3)

        # -- damage intensity
        dNB_H = self.damage_intesity_NB(m0=m0H, nu=v0H, C=C, k=k) 
        dNB_P = self._damage_intesity_bimodal_LF(m0_LF=m0M, m0_HF=m0H, nuP=v0P, C=C, k=k)
        dNB_Q = self._damage_intesity_trimodal_LF(m0_LF=m0L,m0_MF=m0M, m0_HF=m0H, nu_L=v0Q, C=C, k=k)

        d = dNB_H + dNB_P + dNB_Q
        T = 1 / d
        return T

    def _damage_intesity_trimodal_LF(self, m0_LF, m0_MF, m0_HF, nu_L, C, k):
        """Calculates damage intensity for low frequency component of bimodal random process,
        with parameters m0, nuP, C, k, as defined in [2].
        :param m0_LF: [int,float]
            Zeroth spectral moment of low-frequency component [MPa**2].
        :param m0_MF: [int,float]
            Zeroth spectral moment of medium-frequency component [MPa**2].
        :param m0_HF: [int,float]
            Zeroth spectral moment of high-frequency component [MPa**2].    
        :param nu_L: [int,float]
            Frequency of positive slope zero crossing of low frequency component[Hz].
        :param C: [int,float]
            Fatigue strength coefficient [MPa**k].
        :param k : [int,float]
            Fatigue strength exponent [/].
        :return d: float
            Estimated damage intensity of low frequency component.
        """
        # medium-amplitude cycle pdf
        pdf_M = pdf_rayleigh_sum(m0_MF, m0_HF)
        # LF component pdf - rayleigh distributed
        pdf_LF = lambda s: stats.rayleigh.pdf(s/np.sqrt(m0_LF)) / np.sqrt(m0_LF)
        # large-amplitude cycle pdf
        pdf_L = lambda s: np.convolve(pdf_M(s), pdf_LF(s))[0] # large cycle pdf

        S_L = integrate.quad(lambda s: s**k * pdf_L(s), 0, np.inf)[0] 
        d = nu_L * S_L / C
        return d


import numpy as np
from scipy import stats
from scipy import integrate
from scipy.special import gamma
from .narrowband import Narrowband
from ..tools import pdf_rayleigh_sum

class JiaoMoan(Narrowband):
    """Class for fatigue life estimation using frequency domain 
    method by Jiao and Moan [1].
    
    References
    ----------
    [1] Guoyang Jiao and Torgeir Moan. Probabilistic analysis of fatigue due to Gaussian load processes.
        Probabilistic Engineering Mechanics, 5(2):76-83, 1990
    [2] Aleš Zorman and Janko Slavič and Miha Boltežar. 
        Vibration fatigue by spectral methods—A review with open-source support, 
        Mechanical Systems and Signal Processing, 2023, 
        https://doi.org/10.1016/j.ymssp.2023.110149
    
    Example
    -------
    Import modules, define time- and frequency-domain data

    >>> import FLife
    >>> import pyExSi as es
    >>> import numpy as np
    >>> from matplotlib import pyplot as plt
    >>> # time-domain data
    >>> N = 2 ** 16  # number of data points of time signal
    >>> fs = 2048  # sampling frequency [Hz]
    >>> t = np.arange(0, N) / fs  # time vector
    >>> # frequency-domain data
    >>> M = N // 2 + 1  # number of data points of frequency vector
    >>> freq = np.arange(0, M, 1) * fs / N  # frequency vector
    >>> PSD_lower = es.get_psd(freq, 20, 60, variance = 5)  # lower mode of random process
    >>> PSD_higher = es.get_psd(freq, 100, 120, variance = 2)  # higher mode of random process
    >>> PSD = PSD_lower + PSD_higher # bimodal one-sided flat-shaped PSD

    Get Gaussian stationary signal, instantiate SpectralData object and plot PSD

    >>> rg = np.random.default_rng(123) # random generator seed
    >>> x = es.random_gaussian(N, PSD, fs, rg) # Gaussian stationary signal
    >>> sd = FLife.SpectralData(input=x, dt=1/fs) # SpectralData instance
    >>> plt.plot(sd.psd[:,0], sd.psd[:,1]) 
    >>> plt.xlabel('Frequency [Hz]')
    >>> plt.ylabel('PSD')

    Define S-N curve parameters and get fatigue-life estimatate

    >>> C = 1.8e+22  # S-N curve intercept [MPa**k]
    >>> k = 7.3 # S-N curve inverse slope [/]
    >>> jm = FLife.JiaoMoan(sd, PSD_splitting=('userDefinedBands', [80,150]))
    >>> print(f'Fatigue life: {jm.get_life(C,k):.3e} s.')

    Plot segmentated PSD, used in Jiao-Moan method

    >>> lower_band_index, upper_band_index= jm.band_stop_indexes
    >>> plt.plot(sd.psd[:,0], sd.psd[:,1])
    >>> plt.vlines(sd.psd[:,0][lower_band_index], 0, np.max(sd.psd[:,1]), 'k', linestyles='dashed', alpha=.5)
    >>> plt.fill_between(sd.psd[:lower_band_index,0], sd.psd[:lower_band_index,1], 'o', label='lower band', alpha=.2, color='blue')
    >>> plt.vlines(sd.psd[:,0][upper_band_index], 0, np.max(sd.psd[:,1]), 'k', linestyles='dashed', alpha=.5)
    >>> plt.fill_between(sd.psd[lower_band_index:upper_band_index,0], sd.psd[lower_band_index:upper_band_index,1], 'o', label='upper band', alpha=.5, color ='orange')
    >>> plt.xlabel('Frequency [Hz]')
    >>> plt.ylabel('PSD')
    >>> plt.xlim(0,300)
    >>> plt.legend()
    """
    def __init__(self, spectral_data, PSD_splitting = ('equalAreaBands', 2)):
        """Get needed values from reference object.

        :param spectral_data:  Instance of class SpectralData       
        :param PSD_splitting: tuple
                PSD_splitting[0] is PSD spliting method, PSD_splitting[1] is method argument. 
                Splitting methods:

                - 'userDefinedBands', PSD_splitting[1] must be of type list or tupple, with N 
                  elements specifying upper band frequencies of N random processes.
                - 'equalAreaBands', PSD_splitting[1] must be of type int, specifying N random processes.

                Defaults to ('equalAreaBands', 2).
        """
        Narrowband.__init__(self, spectral_data)
        self.PSD_splitting = PSD_splitting
        self.band_stop_indexes = self.spectral_data._get_band_stop_frequency(self.PSD_splitting)
        
    def get_life(self, C, k, approximation = False):
        """Calculate fatigue life with parameters C, k, as defined in [1, 2].

        :param C: [int,float]
            S-N curve intercept [MPa**k].
        :param k: [int,float]
            S-N curve inverse slope [/].
        :param approximation: Boolean
            IF true, approximated PDF of large peaks is used for bimodal random process. 
        :return:
            Estimated fatigue life in seconds.
        :rtype: float
        """
        if len(self.band_stop_indexes) == 1: # narrow-band
            T = self._life_NB(C,k)
        elif len(self.band_stop_indexes) == 2: # bi-modal
            T = self._life_bimodal(C,k,approximation)
        else:
            raise Exception('Specify up to bi-modal random process.')
        return T

    def _life_NB(self, C, k):
        m0H, = self.spectral_data.get_spectral_moments(self.PSD_splitting, moments=[0])[0][0]
        v0H, = self.spectral_data.get_nup(self.PSD_splitting)[0]

        # -- Define expected value of stress range ( int(S^k * p(s)) ) proces H(t), fatigue life
        dNB_H = self.damage_intesity_NB(m0=m0H, nu=v0H, C=C, k=k) 
        T = 1/dNB_H
        return T

    def _life_bimodal(self, C, k, approximation = False):
        # -- spectral moments for each narrowband
        moments = self.spectral_data.get_spectral_moments(self.PSD_splitting, moments=[0,2])
        m0L, m2L = moments[0] #spectral moments for lower band
        m0H, m2H = moments[1] #spectral moments for upper band

        # -- normalized variances
        m0 = np.sum(moments[:, 0])
        m0L_norm = m0L/m0
        m0H_norm = m0H/m0

        # -- Vanmarcke bandwidth parameter
        _, epsV_H = self.spectral_data.get_vanmarcke_parameter(self.PSD_splitting)

        # -- positive slope zero crossing frequency
        v0L, v0H = self.spectral_data.get_nup(self.PSD_splitting)
        v0P = m0L_norm * v0L* np.sqrt(1 + m0H_norm/m0L_norm * (v0H/v0L*epsV_H)**2)

        if approximation:
            # -- bandwidth correction factor
            v0 = 1/(2*np.pi) * np.sqrt((m2L + m2H)/(m0L + m0H))
            rho = v0P/v0 * (m0L_norm**(k/2 + 2) * (1 - np.sqrt(m0H_norm/m0L_norm)) \
                + np.sqrt(np.pi*m0L_norm*m0H_norm) * (k * gamma(k/2 + 1/2))/(gamma(k/2 +1))) \
                    + v0H/v0 * m0H_norm**(k/2)

            # -- damage intensity
            d_nb = self.damage_intesity_NB(m0=m0, nu=v0, C=C, k=k) 
            d = rho * d_nb
        else:
            # -- damage intensity
            dNB_H = self.damage_intesity_NB(m0=m0H, nu=v0H, C=C, k=k) 
            dNB_P = self._damage_intesity_bimodal_LF(m0_LF=m0L, m0_HF=m0H, nuP=v0P, C=C, k=k)
            d = dNB_H + dNB_P

        T = 1 / d
        return T

    def _damage_intesity_bimodal_LF(self, m0_LF, m0_HF, nuP, C, k):
        """Calculates damage intensity for low frequency component of bimodal random process,
        with parameters m0, nuP, C, k, as defined in [2].

        :param m0_LF: [int,float]
            Zeroth spectral moment of low frequency component [MPa**2].
        :param m0_HF: [int,float]
            Zeroth spectral moment of high frequency component [MPa**2].
        :param nuP: [int,float]
            Frequency of positive slope zero crossing of low frequency component[Hz].
        :param C: [int,float]
            Fatigue strength coefficient [MPa**k].
        :param k : [int,float]
            Fatigue strength exponent [/].
        :return d: float
            Estimated damage intensity of low frequency component.
        """
        pdf_P = pdf_rayleigh_sum(m0_LF,m0_HF)
        S_P = integrate.quad(lambda x:  x**k * pdf_P(x), 0, np.inf)[0]
        d = nuP * S_P / C
        return d
            
    def get_PDF(self, s):
        raise Exception(f'Function <get_PDF> is not available for class {self.__class__.__name__:s}.')